program symmetric
!
!  Test a 12th order symmetric multistep method for solving the equation of motion of a single planet circling the Sun.  The
!  Sun is fixed in space.
!
!  John K. Prentice and Agbeli Ameko  
!
!=========================================================================
!
!  This code is part of the Quetzal Computational Associates Fortran 90
!  Benchmark Suite.
!
!  Copyright (c) 1994 by Quetzal Computational Associates.
!
!  Permission to copy and use without fee all or part of the
!  publications or computer codes in this suite is granted provided 
!  that copies are not made or distributed for direct commercial
!  advantage, this copyright notice and the title of the codes
!  or publications and the fact that they originated in this
!  benchmark suite appear, and notice is given that copying is
!  by permission of Quetzal Computational Associates.  For
!  publications that have been published in journals, the name
!  of the publication and date should appear.  To copy otherwise,
!  or to republish, requires a fee and/or specific permission.
!  Authors of codes in this benchmark suite are exempt from the terms of
!  this copyright notice for their codes only.  Address any
!  questions about this copyright notice to:
!
!         John K. Prentice
!         Quetzal Computational Associates
!         3200 Carlisle N.E.
!         Albuquerque, NM   87110-1664
!         USA
!
!         Phone:  505-889-4543
!         Fax:    505-889-4598
!         E-mail: quetzal@aip.org
!
!=========================================================================
!
implicit none
!
!-----------------------------------------------
!   K i n d   P a r a m e t e r s
!-----------------------------------------------
! 
integer, parameter :: LONGreal = selected_real_kind(12,50)  
!
!-----------------------------------------------
!   L o c a l   P a r a m e t e r s
!-----------------------------------------------
!
real (kind = LONGreal), parameter :: zero = 0.0
real (kind = LONGreal), parameter :: point_five = 0.5
real (kind = LONGreal), parameter :: one_point_five = 1.5
real (kind = LONGreal), parameter :: gaussian_constant = 0.01720209895
!
!-----------------------------------------------
!   L o c a l   V a r i a b l e s
!-----------------------------------------------
!
integer :: j
real (kind = LONGreal), dimension(0:12) :: x_position_numerical,y_position_numerical
real (kind = LONGreal), dimension(-1:11) :: x_acceleration,y_acceleration
real (kind = LONGreal), dimension(0:12) :: alpha, beta, gamma
real (kind = LONGreal) ::  time_step, stop_time, radius, time,x_position_exact, y_position_exact, x_dot_exact, y_dot_exact,        &
                         & x_dot_numerical,y_dot_numerical, constant,gravitational_constant,x_error, y_error,exact_total_energy,   &
                         & exact_kinetic_energy, exact_potential_energy, exact_angular_momentum, numerical_total_energy,           &
                         & numerical_kinetic_energy, numerical_potential_energy, numerical_angular_momentum,                       &
                         & exact_velocity_squared
real (kind = LONGreal) ::  numerical_velocity_squared, numerical_radius, numerical_radius_squared, kinetic_energy_error,           &
                         & potential_energy_error, total_energy_error, angular_momentum_error, x_dot_error, y_dot_error,           &
                         & x_alpha_sum, y_alpha_sum, x_beta_sum, y_beta_sum, beta_factor, gamma_factor, x_gamma_sum,               &
                         & y_gamma_sum, radius_error
!
!-----------------------------------------------
!
!   12th order symmetric method coefficients
!
!   Reference:  "Symmetric Multistep Methods for the Numerical
!   Integration of Planetary Orbits", G. D. Quinlan and
!   S. Tremaine, The Astronomical Journal, 100 (1990), page 1695.
!
!   Note!!  The beta below are actually 53,222,400 times
!   the real beta.  This common factor is divided out in the
!   symmetric multistep calculation itself, in order to minimize
!   round-off
!
beta_factor = 53222400.0_LONGreal
alpha(0) = 1.0_LONGreal
beta(0) = 0.0_LONGreal
alpha(1) = -2.0_LONGreal
beta(1) = 90987349.0_LONGreal
alpha(2) = 2.0_LONGreal
beta(2) = -229596838.0_LONGreal
alpha(3) = -1.0_LONGreal
beta(3) = 812627169.0_LONGreal
alpha(4) = 0.0_LONGreal
beta(4) = -1628539944.0_LONGreal
alpha(5) = 0.0_LONGreal
beta(5) = 2714971338.0_LONGreal
alpha(6) = 0.0_LONGreal
beta(6) = -3041896548.0_LONGreal
alpha(7) = 0.0_LONGreal
beta(7) = 2714971338.0_LONGreal
alpha(8) = 0.0_LONGreal
beta(8) = -1628539944.0_LONGreal
alpha(9) = -1.0_LONGreal
beta(9) = 812627169.0_LONGreal
alpha(10) = 2.0_LONGreal
beta(10) = -229596838.0_LONGreal
alpha(11) = -2.0_LONGreal
beta(11) = 90987349.0_LONGreal
alpha(12) = 1.0_LONGreal
beta(12) = 0.0_LONGreal
!
!   12th order Cowell predictor coefficients
!
!   Reference:  "Astronomical Papers Prepared for the Use of the
!   American Ephemeris and Nautical Almanac", C. J. Cohen, E. C.
!   Hubbard, and C. Oesterwinter, 22 (1973), page 20-21.
!
!   Note!!  The gamma below are actually 1,743,565,824,000 times
!   the real gamma.  This common factor is divided out in the
!   Cowell predictor calculation itself, in order to minimize
!   round-off
!
gamma_factor = 1743565824000.0_LONGreal
gamma(0) = 9072652009253.0_LONGreal
gamma(1) = -39726106418680.0_LONGreal
gamma(2) = 140544566352762.0_LONGreal
gamma(3) = -344579280210129.0_LONGreal
gamma(4) = 613137294629235.0_LONGreal
gamma(5) = -811345852376496.0_LONGreal
gamma(6) = 807012356281740.0_LONGreal
gamma(7) = -602852367932304.0_LONGreal
gamma(8) = 333888089374395.0_LONGreal
gamma(9) = -133228219027160.0_LONGreal
gamma(10) = 36262456774618.0_LONGreal
gamma(11) = -6033724094760.0_LONGreal
gamma(12) = 463483373517.0_LONGreal
!
!
!   initialize variables
!
time_step = 0.25_LONGreal
stop_time = 3650000.0_LONGreal
radius = 1.0_LONGreal
time = -time_step
!
print *,' Position solution via 12th order symmetric multistep method'
print *,' Velocity solution via 12th order Cowell predictor method'
print *,'     radius = ', radius, ', time step = ', time_step
!
!   define a constant which is needed later by the exact solution
!
gravitational_constant = gaussian_constant**2
constant = sqrt(gravitational_constant/radius**3)
!
!   initialize the first 12 numerical values using the exact values
!
do j = -1,11
    if (j >= 0) time = time + time_step
    x_position_exact = radius*cos(constant*time)
    y_position_exact = radius*sin(constant*time)
!
    if (j >= 0) then
        x_position_numerical(j) = x_position_exact
        y_position_numerical(j) = y_position_exact
     endif
!
     x_acceleration(j) = -gravitational_constant/radius**3*x_position_exact
     y_acceleration(j) = -gravitational_constant/radius**3*y_position_exact
end do
!
!   compute exact kinetic and potential energies, and the
!   angular momentum.  These values are all divided by the mass
!   of the object.  Since they are conserved, they will never change
!   and hence do not have to be recalculated later.
!
x_dot_exact = -radius*constant*sin(constant*time)
y_dot_exact = radius*constant*cos(constant*time)
exact_velocity_squared = x_dot_exact**2 + y_dot_exact**2
exact_kinetic_energy = point_five*exact_velocity_squared
exact_potential_energy = -gravitational_constant/radius
exact_total_energy = exact_potential_energy + exact_kinetic_energy
exact_angular_momentum = x_position_exact*y_dot_exact-y_position_exact*x_dot_exact
!
!   perform loop over time
!
do while (time <= stop_time)
!
!   advance time step
!
    time = time + time_step
!
!   calculate new acceleration of body at time=time-time_step
!
    numerical_radius_squared = x_position_numerical(11)**2 + y_position_numerical(11)**2
    x_acceleration(11) = -gravitational_constant/numerical_radius_squared**one_point_five*x_position_numerical(11)
    y_acceleration(11) = -gravitational_constant/numerical_radius_squared**one_point_five*y_position_numerical(11)
!
!   numerically solve for the new positions using a 12th order
!   symmetric multistep method.
!
!
!   first sum the first and second terms
!
    x_alpha_sum = dot_product(alpha(0:11:1),x_position_numerical(0:11:1))
    x_beta_sum = dot_product(beta(0:11:1),x_acceleration(0:11:1))
    y_alpha_sum = dot_product(alpha(0:11:1),y_position_numerical(0:11:1))
    y_beta_sum = dot_product(beta(0:11:1),y_acceleration(0:11:1))
!
    x_position_numerical(12) = (-x_alpha_sum) + time_step**2*(x_beta_sum/beta_factor)
    y_position_numerical(12) = (-y_alpha_sum) + time_step**2*(y_beta_sum/beta_factor)
!
!   numerically solve for the new velocities using a 12th order
!   Cowell predictor method.
!
!
!   first sum the gamma terms
!
     x_gamma_sum = dot_product(gamma(0:12:1),x_acceleration(11:(-1):(-1)))
     y_gamma_sum = dot_product(gamma(0:12:1),y_acceleration(11:(-1):(-1)))
!
     x_dot_numerical = (x_position_numerical(11)-x_position_numerical(10))/time_step + time_step*(x_gamma_sum/gamma_factor)
     y_dot_numerical = (y_position_numerical(11)-y_position_numerical(10))/time_step + time_step*(y_gamma_sum/gamma_factor)
!
!   push the stack down one
!
     do j = -1,11
         if (j >= 0) then
             x_position_numerical(j) = x_position_numerical(j+1)
             y_position_numerical(j) = y_position_numerical(j+1)
         endif
!
         if (j < 11) then
             x_acceleration(j) = x_acceleration(j+1)
             y_acceleration(j) = y_acceleration(j+1)
         endif
     end do
!
end do
!
!   print results
!
!
!   first compute energies and angular momenta (add divided by the mass
!   of the object
!
numerical_velocity_squared = x_dot_numerical**2+y_dot_numerical**2
numerical_radius = sqrt(x_position_numerical(12)**2+y_position_numerical(12)**2)
numerical_kinetic_energy = point_five*numerical_velocity_squared
numerical_potential_energy = -gravitational_constant/numerical_radius
numerical_total_energy = numerical_potential_energy+numerical_kinetic_energy
numerical_angular_momentum = x_position_numerical(12)*y_dot_numerical-y_position_numerical(12)*x_dot_numerical
!
!   compute exact results for comparison to the numerical results
!
x_position_exact = radius*cos(constant*time)
y_position_exact = radius*sin(constant*time)
x_dot_exact = -radius*constant*sin(constant*time)
y_dot_exact = radius*constant*cos(constant*time)
!
!   next compute relative errors
!
radius_error = (radius - numerical_radius)/radius
!
if (x_position_exact /= zero) then
    x_error = (x_position_exact - x_position_numerical(12))/x_position_exact
else
    x_error = x_position_numerical(12)
endif
!
if (y_position_exact /= zero) then
    y_error = (y_position_exact - y_position_numerical(12))/y_position_exact
else
    y_error = y_position_numerical(12)
endif
!
if (x_dot_exact /= zero) then
    x_dot_error = (x_dot_exact - x_dot_numerical)/x_dot_exact
else
    x_dot_error = x_dot_numerical
endif
!
if (y_dot_exact /= zero) then
    y_dot_error = (y_dot_exact - y_dot_numerical)/y_dot_exact
else
    y_dot_error = y_dot_numerical
endif
!
kinetic_energy_error = (exact_kinetic_energy - numerical_kinetic_energy)/exact_kinetic_energy
potential_energy_error = (exact_potential_energy - numerical_potential_energy)/exact_potential_energy
total_energy_error = (exact_total_energy - numerical_total_energy)/exact_total_energy
angular_momentum_error = (exact_angular_momentum - numerical_angular_momentum)/exact_angular_momentum
!
!   finally, print out errors
!
print *,' Time = ', time
print *,'    x rel error  = ', x_error, ' y rel error  = ',y_error
print *,'    vx rel error = ', x_dot_error, ' vy rel error = ', y_dot_error
print *,'    KE rel error = ', kinetic_energy_error,' PE rel error = ', potential_energy_error
print *,'    TE rel error = ', total_energy_error,' AM rel error = ', angular_momentum_error
print *,'    numerical radius = ', numerical_radius,' radius rel error = ', radius_error
!
end program symmetric
